//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_UDTSUP_CMANIPBASE_HPP_
#define _LFC1_UDTSUP_CMANIPBASE_HPP_

#include <iostream>

#include "lfc1/udtsup/udtsupexp.hpp"

namespace lfc1 { namespace udtsup {

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/cmanipbase.hpp
///	@brief This class template provides exception/error handling for single argument manipulators of
///		   user-defined types.
///
///	This template was derived from Section 3.2.2.4.1 of the book Standard C++ IOStreams and Locales.  The
///	exception handling sets the stream state and allows the handled exception to propagate if the stream
///	allows it.
///
///	@tparam	T1	Manipulator function argument type.
///
template <typename T1>
class CManipBase
{
public:
	// Type definitions
	typedef void (*TManipFunc)(std::ios_base&, T1);	///< Single argument manipulator signature.

	// Constructors/destructor
	CManipBase(TManipFunc apManipFunc, T1& arArg);
	CManipBase(const CManipBase&) = default;	///< Uses default implementation.
	virtual ~CManipBase() noexcept;

	// Assignment
	CManipBase& operator=(const CManipBase&) = default;	///< Uses default implementation.

private:
	// Instance special operations
	template <typename T2>
	friend void gvDoManip(std::basic_ios<char>& arIos, const CManipBase<T2>& arManipBase);

	// Instance data
	TManipFunc mpManipFunc;
	T1 mvArg;
	std::ios_base::iostate mvIoState;
};

///	@brief This template function creates a default CManipBase object.
///
///	@tparam	T1	Manipulator function argument type.
///
///	@param[in]	apManipFunc	The user-defined manipulator function.
///	@param[in]	arArg		The user-defined manipulator function argument.
///
template <typename T1>
CManipBase<T1>::CManipBase(TManipFunc apManipFunc, T1& arArg)
:	mpManipFunc(apManipFunc),
	mvArg(arArg),
	mvIoState(std::ios_base::goodbit)
{
}

///	@brief This template function destroys a CManipBase object.
///
///	@tparam	T1	Manipulator function argument type.
///
template <typename T1>
CManipBase<T1>::~CManipBase() noexcept
{
}

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/cmanipobase.hpp
///	@brief This template function calls the user-defined manipulator function and handles exceptions from it.
///
///	@tparam	T2	The user-defined manipulator function argument type.
///
///	@param	arIos		The stream's state object.
///	@param	arManipBase	The user-defined manipulator function container.
///
template <typename T2>
void gvDoManip(std::basic_ios<char>& arIos, const CManipBase<T2>& arManipBase)
{
	if (arManipBase.mvIoState != std::ios_base::goodbit)
	{
		arIos.setstate(arManipBase.mvIoState);
	}
	else
	{
		auto vIoState = arManipBase.mvIoState;
		try
		{
			(*(arManipBase.mpManipFunc))(arIos, arManipBase.mvArg);
		}
		catch (std::bad_alloc&)
		{
			// Memory allocation failure sets the badbit flag on the manipulator.
			vIoState |= std::ios_base::badbit;
			auto vMask = arIos.exceptions();
			if ((vMask & std::ios_base::failbit) != 0 && (vMask & std::ios_base::badbit) == 0)
			{
				arIos.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::badbit) != 0)
			{
				try
				{
					arIos.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}
		catch (...)
		{
			// Any other failure sets the failbit flag on the manipulator.
			vIoState |= std::ios_base::failbit;
			auto vMask = arIos.exceptions();
			if ((vMask & std::ios_base::badbit) != 0 && (vIoState & std::ios_base::badbit) != 0)
			{
				arIos.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::failbit) != 0)
			{
				try
				{
					arIos.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}

		if (vIoState != std::ios_base::goodbit)
		{
			arIos.setstate(vIoState);
		}
	}
}

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/cmanipobase.hpp
///	@brief This template function invokes a user-defined manipulator for an input stream.
///
///	@tparam	T3	The user-defined manipulator function argument type.
///
///	@param[in]	arStream	The source stream.
///	@param[out]	arManipBase	The CManipBase object.
///
///	@return	The source stream.
///
template <typename T3>
std::istream& operator>>(std::istream& arStream, const CManipBase<T3>& arManipBase)
{
	// Check if the input stream is still in a good state.
	// No point in doing anything if it is not in a good state.
	if (arStream.good())
	{
		gvDoManip<T3>(arStream, arManipBase);
	}
	return arStream;
}

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/cmanipobase.hpp
///	@brief This template function invokes a user-defined manipulator for an output stream.
///
///	@tparam	T3	The user-defined manipulator function argument type.
///
///	@param[in]	arStream	The destination stream.
///	@param[in]	arManipBase	The CManipBase object.
///
///	@return	The destination stream.
///
template <typename T3>
std::ostream& operator<<(std::ostream& arStream, const CManipBase<T3>& arManipBase)
{
	// Check if the output stream is still in a good state.
	// No point doing anything if it is not in a good state.
	if (arStream.good())
	{
		gvDoManip<T3>(arStream, arManipBase);
	}
	return arStream;
}

} /* namespace udtsup */ } /* namespace lfc1 */

#endif // _LFC1_UDTSUP_CMANIPBASE_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
