//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_UDTSUP_OUTPUT_HPP_
#define _LFC1_UDTSUP_OUTPUT_HPP_

#include <iostream>
#include <type_traits>

#include "lfc1/udtsup/udtsupexp.hpp"

namespace lfc1 { namespace udtsup {

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/output.hpp
///	@brief This template function is meant to be the only function called by the inserter of a user-defined
///		   type.  This function provides exception/error handling for the extractor of a user-defined type.
///
///	This template was derived from Section 3.1.5 of the book Standard C++ IOStreams and Locales.  The
///	exception handling sets the stream state and allows the handled exception to propagate if the stream
///	allows it.
///
///	The user-defined type is expected to contain a function with the following signature:
///
///		@code
///		std::ios_base::iostate mvOutput(std::istream&)
///		@endcode
///
///	The mvOutput() function can call inserters for built-in types thus this function does not instantiate a
///	stream sentry object.
///
///	@tparam	UDT	The user-defined type.
///
///	@param[in]	arStream	The destination stream.
///	@param[out]	arUdt		The user-defined type.
///
///	@return The destination stream.
///
template <typename UDT>
std::ostream& gvOutput(std::ostream& arStream, const UDT& arUdt)
{
	// UDT must be a user-defined type which has member function mvOutput().
	static_assert
	(
		std::is_class<UDT>::value,
		"User-defined type must be a class"
	);
	static_assert
	(
		std::is_member_function_pointer<std::ios_base::iostate (UDT::*)(std::ostream&)>::value,
		"User-defined type must have member function mvOutput()"
	);

	// Check if the output stream is still in a good state.
	// No point doing anything if it is not in a good state.
	if (arStream.good())
	{
		// Initialize a local stream state.
		auto vIoState = std::ios_base::goodbit;

		// Send the output to the destination stream.
		try
		{
			vIoState = arUdt.mvOutput(arStream);
		}
		catch (std::ios_base::failure&)
		{
			throw;
		}
		catch (std::bad_alloc&)
		{
			// Memory allocation failure sets the badbit flag on the destination stream.
			vIoState |= std::ios_base::badbit;
			auto vMask = arStream.exceptions();
			if ((vMask & std::ios_base::failbit) != 0 && (vMask & std::ios_base::badbit) == 0)
			{
				arStream.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::badbit) != 0)
			{
				try
				{
					arStream.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}
		catch (...)
		{
			// Any other failure sets the failbit flag on the source stream.
			vIoState |= std::ios_base::failbit;
			auto vMask = arStream.exceptions();
			if ((vMask & std::ios_base::badbit) != 0 && (vIoState & std::ios_base::badbit) != 0)
			{
				arStream.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::failbit) != 0)
			{
				try
				{
					arStream.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}

		// If the local stream state has changed, save it to the destination stream.
		if (vIoState != std::ios_base::goodbit)
		{
			arStream.setstate(vIoState);
		}
	}
	return arStream;
}

///	@ingroup LFC1_UDTSUP
///	@headerfile lfc1/udtsup/output.hpp
///	@brief This template function is meant to be the only function called by the inserter of a user-defined
///		   type.  This function provides exception/error handling for the extractor of a user-defined type.
///
///	This template was derived from Section 3.1.5 of the book Standard C++ IOStreams and Locales.  The
///	exception handling sets the stream state and allows the handled exception to propagate if the stream
///	allows it.
///
///	The user-defined type is expected to contain a function with the following signature:
///
///		@code
///		std::ios_base::iostate mvOutput(std::istream&)
///		@endcode
///
///	The mvOutput() function must not call inserters for built-in types because this function instantiates a
///	stream sentry object.
///
///	@tparam	UDT	The user-defined type.
///
///	@param[in]	arStream	The destination stream.
///	@param[out]	arUdt		The user-defined type.
///
///	@return The destination stream.
///
template <typename UDT>
std::ostream& gvOutputGuarded(std::ostream& arStream, const UDT& arUdt)
{
	// UDT must be a user-defined type which has member function mvOutput().
	static_assert
	(
		std::is_class<UDT>::value,
		"User-defined type must be a class"
	);
	static_assert
	(
		std::is_member_function_pointer<std::ios_base::iostate (UDT::*)(std::ostream&)>::value,
		"User-defined type must have member function mvOutput()"
	);

	// Check if the output stream is still in a good state.
	// No point doing anything if it is not in a good state.
	if (arStream.good())
	{
		// Initialize a local stream state.
		auto vIoState = std::ios_base::goodbit;

		// Send the output to the destination stream.
		try
		{
			typename std::ostream::sentry vGuard(arStream);
			if (vGuard)
			{
				vIoState = arUdt.mvOutput(arStream);
				arStream.width(0);
			}
		}
		catch (std::bad_alloc&)
		{
			// Memory allocation failure sets the badbit flag on the destination stream.
			vIoState |= std::ios_base::badbit;
			auto vMask = arStream.exceptions();
			if ((vMask & std::ios_base::failbit) != 0 && (vMask & std::ios_base::badbit) == 0)
			{
				arStream.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::badbit) != 0)
			{
				try
				{
					arStream.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}
		catch (...)
		{
			// Any other failure sets the failbit flag on the source stream.
			vIoState |= std::ios_base::failbit;
			auto vMask = arStream.exceptions();
			if ((vMask & std::ios_base::badbit) != 0 && (vIoState & std::ios_base::badbit) != 0)
			{
				arStream.setstate(vIoState);
			}
			else if ((vMask & std::ios_base::failbit) != 0)
			{
				try
				{
					arStream.setstate(vIoState);
				}
				catch (std::ios_base::failure&)
				{
				}
				throw;
			}
		}

		// If the local stream state has changed, save it to the destination stream.
		if (vIoState != std::ios_base::goodbit)
		{
			arStream.setstate(vIoState);
		}
	}
	return arStream;
}

} /* namespace udtsup */ } /* namespace lfc1 */

#endif // _LFC1_UDTSUP_OUTPUT_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
